#include "ch32_pd.h"

#include "ch32l103.h"
#include "ch32l103_usbpd.h"

#include <string.h>

#include "usbc_pd.h"
#include "usbc_tc.h"

#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define MAX(a, b) ((a) > (b) ? (a) : (b))

#define BIT(n) (1 << (n))

struct ucpd_ch32_data
{
    enum tc_cc_polarity polarity;
    bool rx_sop_prime;
    __attribute__((aligned(4))) uint8_t buffer[512];
    bool tx_good_crc;
    bool tx_need_crc;
    bool tx_valid;
    struct pd_message rx_msg;
    bool rx_msg_valid;
};

static struct ucpd_ch32_data ucpd_ch32_data;

static EventGroupHandle_t pd_event_group;
static StaticEventGroup_t pd_event_group_buffer;

static bool ucpd_ch32_is_good_crc(union pd_header header)
{
    return header.number_of_data_objects == 0 &&
           header.extended == 0 &&
           header.message_type == PD_CTRL_GOOD_CRC;
}

static void ucpd_ch32_end_tx(void)
{
    struct ucpd_ch32_data *data = &ucpd_ch32_data;
    if (data->polarity == TC_POLARITY_CC1)
    {
        USBPD->PORT_CC1 &= ~CC_LVE;
    }
    else
    { // TC_POLARITY_CC2
        USBPD->PORT_CC2 &= ~CC_LVE;
    }
}

static void ucpd_ch32_start_rx(void)
{
    USBPD->CONFIG |= PD_ALL_CLR;
    USBPD->CONFIG &= ~PD_ALL_CLR;
    USBPD->CONTROL &= ~PD_TX_EN;
    USBPD->BMC_CLK_CNT = UPD_TMR_RX_96M;
    USBPD->CONTROL |= BMC_START;
}

int ucpd_ch32_check_polarity(void)
{
    enum tc_cc_polarity polarity = TC_POLARITY_CC1;

    USBPD->PORT_CC1 &= ~(CC_CE | PA_CC_AI);
    USBPD->PORT_CC1 |= CC_CMP_22;
    portYIELD();
    if (USBPD->PORT_CC1 & PA_CC_AI)
        polarity = TC_POLARITY_CC1;

    USBPD->PORT_CC2 &= ~(CC_CE | PA_CC_AI);
    USBPD->PORT_CC2 |= CC_CMP_22;
    portYIELD();
    if (USBPD->PORT_CC2 & PA_CC_AI)
        polarity = TC_POLARITY_CC2;

    return polarity;
}

static int ucpd_ch32_set_cc_polarity(enum tc_cc_polarity polarity)
{
    struct ucpd_ch32_data *data = &ucpd_ch32_data;
    data->polarity = polarity;
    if (data->polarity)
    {
        USBPD->CONFIG = USBPD->CONFIG & ~CC_SEL;
    }
    else
    { // TC_POLARITY_CC2
        USBPD->CONFIG = USBPD->CONFIG | CC_SEL;
    }
    ucpd_ch32_start_rx(); // don't start receive until after polarity set
    return 0;
}

void ucpd_ch32_start_tx(enum pd_packet_type type, union pd_header header,
                        uint8_t *buffer, uint32_t len)
{
    struct ucpd_ch32_data *data = &ucpd_ch32_data;
    // setup packet
    len = MIN(len, sizeof(data->buffer) - 2);
    switch (type)
    {
    case PD_PACKET_SOP:
        USBPD->TX_SEL = UPD_SOP0;
        data->tx_need_crc = true;
        break;
    case PD_PACKET_SOP_PRIME:
        USBPD->TX_SEL = UPD_SOP1;
        data->tx_need_crc = true;
        break;
    case PD_PACKET_PRIME_PRIME:
        USBPD->TX_SEL = UPD_SOP2;
        data->tx_need_crc = true;
        break;
    case PD_PACKET_TX_HARD_RESET:
        USBPD->TX_SEL = UPD_HARD_RESET;
        data->tx_need_crc = false;
        break;
    case PD_PACKET_CABLE_RESET:
        USBPD->TX_SEL = UPD_CABLE_RESET;
        data->tx_need_crc = false;
        break;
    default:
        printf("unsupported packet type %d, default to SOP", type);
        data->tx_need_crc = true;
        USBPD->TX_SEL = UPD_SOP0;
        break;
    }
    USBPD->BMC_TX_SZ = 2 + len;
    data->buffer[0] = (header.raw_value >> 0) & 0xFF;
    data->buffer[1] = (header.raw_value >> 8) & 0xFF;
    memcpy(&data->buffer[2], buffer, len);
    if (!ucpd_ch32_is_good_crc(header))
    {
        data->tx_valid = true;
    }

    // send packet
    if (data->polarity == TC_POLARITY_CC1)
    {
        USBPD->PORT_CC1 |= CC_LVE;
    }
    else
    { // TC_POLARITY_CC2
        USBPD->PORT_CC2 |= CC_LVE;
    }
    USBPD->BMC_CLK_CNT = UPD_TMR_TX_96M;
    USBPD->CONTROL |= PD_TX_EN;
    USBPD->STATUS &= BMC_AUX_INVALID;
    USBPD->CONTROL |= BMC_START;
}

int ucpd_rx_pending(void)
{
    return ucpd_ch32_data.rx_msg_valid;
}

uint8_t ucpd_read_message(pd_msg *msg)
{
    struct ucpd_ch32_data *data = &ucpd_ch32_data;
    if (!data->rx_msg_valid)
    {
        return 1;
    }
    msg->hdr = data->rx_msg.header.raw_value;
    memcpy(msg->obj, data->rx_msg.data, data->rx_msg.len);
    return 0;
}

void ucpd_send_message(const pd_msg *msg)
{
    struct ucpd_ch32_data *data = &ucpd_ch32_data;
    union pd_header header;
    int retries = 3;
    int msg_len = 4 * PD_NUMOBJ_GET(msg);

    header.raw_value = msg->hdr;
    do {
        ucpd_ch32_start_tx(PD_PACKET_SOP, header, msg->obj, msg_len);
        vTaskDelay(pdMS_TO_TICKS(1));
        if (!data->tx_valid) {
            xEventGroupSetBits(pd_event_group, PD_MSG_SENT_EVENT);
            return;
        }
    } while (--retries);
    xEventGroupSetBits(pd_event_group, PD_MSG_SEND_FAILED_EVENT);
}

int ucpd_ch32_init(void)
{
    GPIO_InitTypeDef GPIO_InitStructure = {0};

    RCC_PB2PeriphClockCmd(RCC_PB2Periph_GPIOB, ENABLE);
    RCC_PB2PeriphClockCmd(RCC_PB2Periph_AFIO, ENABLE);
    RCC_HBPeriphClockCmd(RCC_HBPeriph_USBPD, ENABLE);
    GPIO_InitStructure.GPIO_Pin = GPIO_Pin_6 | GPIO_Pin_7;
    GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
    GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN_FLOATING;
    GPIO_Init(GPIOB, &GPIO_InitStructure);
    AFIO->CR |= USBPD_IN_HVT;

    ucpd_ch32_data.polarity = TC_POLARITY_CC1;
    ucpd_ch32_data.rx_sop_prime = false;
    ucpd_ch32_data.tx_good_crc = false;
    ucpd_ch32_data.tx_need_crc = false;
    ucpd_ch32_data.tx_valid = false;
    ucpd_ch32_data.rx_msg_valid = false;

    pd_event_group = xEventGroupCreateStatic(&pd_event_group_buffer);

    USBPD->STATUS = (BUF_ERR | IF_RX_BIT | IF_RX_BYTE | IF_RX_ACT | IF_RX_RESET | IF_TX_END);
    USBPD->DMA = (uint32_t)ucpd_ch32_data.buffer;
    USBPD->CONFIG = IE_RX_ACT | IE_RX_RESET | IE_TX_END | PD_DMA_EN | PD_FILT_ED;

    ucpd_ch32_set_cc_polarity(ucpd_ch32_check_polarity());
    // ucpd_ch32_update_rp();

    NVIC_EnableIRQ(USBPD_IRQn);

    return 0;
}

void USBPD_IRQHandler(void) __attribute__((interrupt()));
void USBPD_IRQHandler(void)
{
    struct ucpd_ch32_data *data = &ucpd_ch32_data;
    BaseType_t xHigherPriorityTaskWoken = pdFALSE;
    uint8_t status = USBPD->STATUS;

    printf("USBPD_IRQHandler status: %d\r\n", status);

    if (status & IF_RX_ACT)
    {
        uint16_t byte_count = USBPD->BMC_BYTE_CNT;
        if ((byte_count < (2 + 4)) || (byte_count > (PD_MAX_EXTENDED_MSG_LEN + 2 + 4)))
        {
            USBPD->STATUS |= IF_RX_ACT;
            return;
        }
        else
        {
            if (data->rx_msg_valid)
            {
                printf("message overrun, replacing old packet\r\n");
            }
            switch (status & MASK_PD_STAT)
            {
            case PD_RX_SOP0:
                data->rx_msg.type = PD_PACKET_SOP;
                break;
            case PD_RX_SOP1_HRST:
                data->rx_msg.type = PD_PACKET_SOP_PRIME;
                break;
            case PD_RX_SOP2_CRST:
                data->rx_msg.type = PD_PACKET_PRIME_PRIME;
                break;
            default:
                data->rx_msg.type = PD_PACKET_MSG_INVALID;
                break;
            }
            data->rx_msg.header.raw_value = data->buffer[0] | (data->buffer[1] << 8);
            data->rx_msg.len = byte_count - 2 - 4; // remove header and CRC
            memcpy(data->rx_msg.data, &data->buffer[2], data->rx_msg.len);

            if (!ucpd_ch32_is_good_crc(data->rx_msg.header) &&
                (data->rx_msg.type == PD_PACKET_SOP || data->rx_sop_prime))
            {
                union pd_header header = {
                    .message_type = PD_CTRL_GOOD_CRC,
                    .port_data_role = 0,
                    .specification_revision = data->rx_msg.header.specification_revision,
                    .port_power_role = 0,
                    .message_id = data->rx_msg.header.message_id,
                    .number_of_data_objects = 0,
                    .extended = 0,
                };
                if (data->rx_msg.type == PD_PACKET_SOP)
                {
                    header.port_power_role = TC_ROLE_SINK;
                    header.port_data_role = TC_ROLE_UFP;
                }
                data->tx_good_crc = true;
                ucpd_ch32_start_tx(data->rx_msg.type, header, NULL, 0);
                // hold lock until GoodCRC sent
            }
            else
            {
                if (data->tx_valid && ucpd_ch32_is_good_crc(data->rx_msg.header))
                {
                    data->tx_valid = false;
                }
                data->rx_msg_valid = true;
                xEventGroupSetBitsFromISR(pd_event_group, PD_MSG_RECEIVED_EVENT, &xHigherPriorityTaskWoken);
                portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
                // no GoodCRC needed, send to stack immediately
            }
        }
        USBPD->STATUS |= IF_RX_ACT;
    }

    if (status & IF_TX_END)
    {
        ucpd_ch32_end_tx();
        ucpd_ch32_start_rx();
        if (data->tx_good_crc)
        {
            data->tx_valid = false;
            data->rx_msg_valid = true;
            xEventGroupSetBitsFromISR(pd_event_group, PD_MSG_RECEIVED_EVENT, &xHigherPriorityTaskWoken);
            portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
        }
        else if (!data->tx_need_crc)
        {
            data->tx_valid = false;
            xEventGroupSetBitsFromISR(pd_event_group, PD_MSG_SENT_EVENT, &xHigherPriorityTaskWoken);
            portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
        }
        USBPD->STATUS |= IF_TX_END;
    }

    if (status & IF_RX_RESET)
    {
        USBPD->STATUS |= IF_RX_RESET;
    }
}

EventBits_t ucpd_get_event_bits(void)
{
    return xEventGroupGetBits(pd_event_group);
}

EventBits_t ucpd_clear_event_bits(EventBits_t bits)
{
    return xEventGroupClearBits(pd_event_group, bits);
}
